/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * Implementation of logging facilities.
 *
 * Europagate, 1994-1995.
 *
 * $Log: gw-log.c,v $
 * Revision 1.13  1995/12/01 12:41:03  adam
 * Bug fix.
 *
 * Revision 1.12  1995/12/01  12:24:17  adam
 * New function: gw_log_mask_str.
 *
 * Revision 1.11  1995/11/09  09:54:28  adam
 * More readable logging format.
 *
 * Revision 1.10  1995/05/16  09:40:48  adam
 * LICENSE.
 *
 * Revision 1.9  1995/04/19  12:12:06  adam
 * Resource system uses only one log debug level.
 *
 * Revision 1.8  1995/04/17  09:36:16  adam
 * Minor changes.
 *
 * Revision 1.7  1995/04/10  13:20:25  adam
 * Use gettimeofday(2) instead of time(2) to get log time in milliseconds.
 *
 * Revision 1.6  1995/03/28  08:01:51  adam
 * Bug fix: GW_LOG_ERRNO.
 *
 * Revision 1.5  1995/03/27  12:51:10  adam
 * New log level in use: GW_LOG_ERRNO.
 *
 * Revision 1.4  1995/02/23  08:32:22  adam
 * Changed header.
 *
 * Revision 1.2  1995/02/17  17:06:56  adam
 * Remove everything before '/' in app_name. Use compact date format.
 *
 * Revision 1.1.1.1  1995/02/09  17:27:12  adam
 * Initial version of email gateway under CVS control.
 *
 * Initial:       Dec  7, 94 (Adam Dickmeiss)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>

#include <gw-log.h>

static char *app_name = NULL;
static unsigned level = GW_LOG_DEFAULT;
static int session = 0;

struct file_mask {
    unsigned mask;           /* level mask for this file entry */
    int fd;                  /* file descriptor for this file */
    char *fname;             /* name of file ("" if stdout) */
    struct file_mask *next;  /* next file in chain */
};

struct file_mask *file_mask_list = NULL;

char *gw_strdup (const char *s)
{
    char *n = malloc (strlen(s)+1);
    if (n)
        strcpy (n, s);
    return n;
}

void gw_log_init (const char *app_name_a)
{
    struct file_mask *list, *list1;
    const char *cp;

    if ((cp = strrchr (app_name_a, '/')))
        app_name = gw_strdup (cp+1);
    else
        app_name = gw_strdup (app_name_a);
    level = GW_LOG_DEFAULT;
    session = 0;

    /* clean up all output file masks... */
    for (list = file_mask_list; list; list = list1)
    {
        if (list->fd > 2)                  /* avoid closing stdout/stderr */
            close (list->fd);              
        free (list->fname);
        list1 = list->next;
        free (list);
    }
    file_mask_list = NULL;
}

void gw_log_level (unsigned level_a)
{
    level = level_a;
}

void gw_log_session (int session_a)
{
    session = session_a;
}

int gw_log_file (unsigned level_a, const char *fname_a)
{
    struct file_mask **listp, *new_file_mask;
    listp = &file_mask_list;

    /* go through file mask list and close files already associated */
    /* on new level */
    while (*listp)
    {
        if (!((*listp)->mask &= ~level_a)) /* any levels left on this one? */
        {
            if ((*listp)->fd > 2)          /* close if not stdout/stderr */
                close ((*listp)->fd);      
            free ((*listp)->fname);
            *listp = (*listp)->next;
        }
        listp = &(*listp)->next;
    }
    if (!fname_a)                          /* stderr? */
        return 0;                          /* stderr doesn't appear on list */
    new_file_mask = malloc (sizeof(*new_file_mask));
    if (!new_file_mask)
        return -1;
    new_file_mask->mask = level_a;
    new_file_mask->next = file_mask_list;
    file_mask_list = new_file_mask;
    if (!(file_mask_list->fname = gw_strdup (fname_a)))
        return -1;
    if (*fname_a == '\0')                  /* stdout? */
        new_file_mask->fd = 1;            
    else                                   /* no, open as usual */
    {
        new_file_mask->fd = open (fname_a, O_WRONLY|O_CREAT|O_APPEND, 0666);
        if (new_file_mask->fd == -1)
            return -1;
    }
    return 0;
}

int gw_log (unsigned level_a, const char *event_type, const char *format, ...)
{
    static char emit_str[4096];
    struct file_mask *list;
    struct timeval tv;
    struct timezone tz;
    va_list ap;
    unsigned e_level = level_a & level;
    int count;
    int err = 0;
    struct tm tms;
    char *cp;

    if (!e_level)                          /* any effective level(s)? */      
        return 0;

    va_start (ap, format);
    gettimeofday (&tv, &tz);

    memcpy (&tms, localtime (&tv.tv_sec), sizeof(tms));
    sprintf (emit_str, "%s %d %02d%02d%02d %02d%02d%02d %03d %d %s ",
             app_name, session,
             tms.tm_year, 1+tms.tm_mon, tms.tm_mday,
             tms.tm_hour, tms.tm_min, tms.tm_sec,
             (int) (tv.tv_usec/1000),
             e_level, event_type);
    if ((cp = strchr (emit_str, '\n')))    /* remove \n from ctime-str */
        *cp = ' ';
    count = strlen (emit_str);
    vsprintf (emit_str+count, format, ap);
    if (level_a & GW_LOG_ERRNO)
    {
        strcat (emit_str, ": ");
        strcat (emit_str, strerror (errno));
    }
    strcat (emit_str, "\n");
    count = strlen (emit_str);

    /* go through file mask list... */
    for (list = file_mask_list; list; list = list->next)
        if (list->mask & e_level)          /* output this one? */
        {
            e_level &= ~list->mask;        /* remove from effective level */
            if (write (list->fd, emit_str, count) != count)
                err = 1;
        }
    if (e_level)                           /* bits left on effective level? */
        write (2, emit_str, count);
    va_end (ap);
    return err;
}

static struct {
    int mask;
    char *name;
} mask_names[] =
{
    { GW_LOG_ALL,       "all"    },
    { GW_LOG_DEFAULT,   "default"},
    { GW_LOG_DEFAULT,   "def"    },
    { GW_LOG_FATAL,     "fatal"  },
    { GW_LOG_WARN,      "warn"   },
    { GW_LOG_ACCT,      "acct"   },
    { GW_LOG_STAT,      "stat"   },
    { GW_LOG_NOTICE,    "notice"   },
    { GW_LOG_DEBUG,     "debug"  },
    { GW_LOG_DEBUGN(0), "debug0" },
    { GW_LOG_DEBUGN(1), "debug1" },
    { GW_LOG_DEBUGN(2), "debug2" },
    { GW_LOG_DEBUGN(3), "debug3" },
    { GW_LOG_DEBUGN(4), "debug4" },
    { GW_LOG_DEBUGN(5), "debug5" },
    { GW_LOG_DEBUGN(6), "debug6" },
    { GW_LOG_DEBUGN(7), "debug7" },
    { GW_LOG_DEBUGN(8), "debug8" },
    { GW_LOG_DEBUGN(8), "debug9" },
    { 0,                "none"   },
    { 0, NULL }
};  

unsigned gw_log_mask_str (const char *str)
{
    const char *p;
    int i;
    unsigned level = GW_LOG_DEFAULT;

    while (*str)
    {
        for (p = str; *p && *p != ','; p++)
            ;
        if (*str == '-' || isdigit(*str))
            level = atoi (str);
        else
            for (i = 0; mask_names[i].name; i++)
                if (strlen (mask_names[i].name) == p-str &&
                    memcmp (mask_names[i].name, str, p-str) == 0)
                {
                    if (mask_names[i].mask)
                        level |= mask_names[i].mask;
                    else
                        level = 0;
                }
        if (*p == ',')
            p++;
        str = p;
    }
    return level;
}
